package Zaphod::Editor::Cursor;
use strict;

=head1 NAME

Zaphod::Editor::Cursor - encapsulates the cursor position

=cut

use vars qw($VERSION);
$VERSION = '0.02';

use base 'Class::Accessor';
use Scalar::Util qw(weaken);
use Data::Dumper;

__PACKAGE__->mk_accessors(qw(column line offset editor));

=head2 C<< __PACKAGE__->new ARGS >>

Constructs a new insertion point object

=cut

sub new {
    my ($class,%args) = @_;
    
    weaken $args{editor};
    my $self = $class->SUPER::new(\%args);

    if (defined $args{offset} and not defined $args{line}) {
        # calculate line/column from offset
        $self->recalc_line_from_ofs();

    } elsif (defined $args{line} and not defined $args{offset}) {
        # calculate offset from line/column
        my $r = $self->editor->docref;
        if ($$r =~ /^(?:.*?\Q\n\E){$args{line}}.{$args{column}}/) {
            $self->offset(pos $$r);
        } else {
            warn "Position ($args{column};$args{line}) outside of document";
            $self->start_of_document;
        }
    } else {
        # no consistency check is made here for quick cloning
        #$args{line}   ||= 0;
        #$args{column} ||= 0;
        #$args{offset} ||= 0;
    };

    $self
}

sub recalc_line_from_ofs {
    my ($self) = @_;
    my $e = $self->editor;
    my $r = $e->docref;
    my $offset = $self->offset;
    $self->line(substr $r, 0, $offset =~ /\n/g);
    $self->column($offset - rindex( substr( $r, 0, $offset ), "\n")+1);
}

=head2 C<< $position->copy ARGS >>

Returns a shallow copy of the position
and overwrites the values with C<ARGS>
where they are given

=cut

sub copy {
    my ($self,%args) = @_;
    %args = (%$self, %args);

    # Do we need to re-weaken $args{editor} here?

    (ref $self)->new(%args)
}

=head2 C<< $p->as_string >>

Returns the position as a convenient string representation

=cut

sub as_string {
    my ($self) = @_;
    sprintf "(c %s;l %s)", $self->pos
}

=head2 C<< $p->pos >>

Returns the column and the line as a list of two integers

=cut

sub pos {
    my ($self) = @_;
    ($self->column, $self->line)
};

=head2 C<< move_left COUNT >>

Moves the cursor position COUNT elements
to the left, wrapping lines if necessary.

Currently, it clamps the value to 0 instead.

(this should be configurable instead of
hard-coded)

In void context, modifies the object in-place.
In other context, a modified copy is returned
and the original object remains as-is.

=cut

sub move_left {
    my ($self,$count) = @_;
    
    my $obj = defined wantarray ? $_[0]->copy() : $_[0];

    my $ofs = $self->offset;
    $ofs -= $count;
    $ofs = 0 if $ofs < 0;
    $self->offset($ofs);

    # Premature optimization is the root of all evil
    if (my $col = $self->column > $count) {
        $self->column($col-$count)
    } else {
        $self->recalc_line_from_ofs();
    };

    $obj
}

=head2 C<< $c->move_right COUNT >>

Moves the cursor position COUNT elements
to the right, wrapping lines if necessary.

Currently, it clamps the value to C<length $line> instead.

(this should be configurable instead of
hard-coded)

In void context, modifies the object in-place.
In other context, a modified copy is returned
and the original object remains as-is.

=cut

sub move_right {
    my ($self,$count) = @_;
    $count = 1 unless defined $count;

    my $obj = defined wantarray ? $_[0]->copy() : $_[0];

    my $col = $obj->column +$count;

    # This is quite inefficient
    my $offset = $obj->offset + $count;
    $obj->recalc_line_from_ofs();

    $obj
}

sub move_up {
    my ($self,$count) = @_;
    $count = 1 unless defined $count;

    my $obj = defined wantarray ? $_[0]->copy() : $_[0];

    my $offset = $obj->offset;
    my $r = $obj->editor->docref;
    
    if ($count) {
        my $col = $obj->column;
        $offset -= $col;
        while ($count) {
            $offset = rindex "\n", substr $$r, 0, $offset - 1;
            $count--;
        };
        
        # This is not entirely correct - 
        # lines shorter than $col shouldn't wrap!
        $offset += $col;
        $obj->offset($offset);
        $obj->recalc_line_from_ofs();
    };

    $obj
}

sub move_down {
    my ($self,$count) = @_;
    $count = 1 unless defined $count;

    my $obj = defined wantarray ? $_[0]->copy() : $_[0];

    my $r = $obj->editor->docref;
    
    my $offset = $obj->offset;
    if ($count) {
        my $col = $obj->column;
        $offset -= $col;
        #my $nl = $obj->editor->newline;
        while ($count) {
            $offset += index "\n", substr $$r, $offset;
            $count--;
        };
        
        # This is not entirely correct - 
        # lines shorter than $col shouldn't wrap!
        $offset += $col;
        $obj->offset($offset);
        $obj->recalc_line_from_ofs();
    };

    $obj
}

sub start_of_line {
    my ($self) = @_;
    my $obj = defined wantarray ? $_[0]->copy() : $_[0];

    $obj->offset( $obj->offset - $obj->column );
    $obj->column(0);

    $obj
};

sub start_of_text_on_line {
    my ($self) = @_;
    my $obj = defined wantarray ? $_[0]->copy() : $_[0];
    
    $obj->start_of_line;
    
    my $col = 0;
    if ($self->editor->current_line =~ /\S/) {
        $col = pos $self->editor->current_line;
    };
    $obj->column($col);
    $obj->offset($obj->offset + $col);

    $obj
};

sub end_of_line {
    my ($self) = @_;
    my $obj = defined wantarray ? $_[0]->copy() : $_[0];

    $obj->start_of_line;
    $obj->column(length($self->editor->current_line));
    $obj->offset($obj->offset + $obj->column);
    
    $obj
};

sub start_of_document {
    my ($self) = @_;
    my $obj = defined wantarray ? $_[0]->copy() : $_[0];

    $obj->column(0);
    $obj->line(0);
    $obj->offset(0);

    $obj
};

sub end_of_document {
    my ($self) = @_;
    my $obj = defined wantarray ? $_[0]->copy() : $_[0];

    $obj->offset( length ${ $obj->editor->docref });

    $obj->recalc_line_from_ofs();
    $obj
};

1;

=head1 AUTHOR

Written by Max Maischein E<lt>corion@cpan.orgE<gt>

Copyright Max Maischein 2005,2006; All Rights Reserved.

This program is free software; you can redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

L<Zaphod>

=cut
